/*
 *  Copyright (c) 1992 John E. Davis  (davis@amy.tch.harvard.edu)
 *  All Rights Reserved.
 */
#include <stdio.h>
#include <string.h>
#ifndef msdos
#ifndef VMS
#include <malloc.h>
#endif  
#else
#include <alloc.h>
#endif
#include "buffer.h"
#include "keymap.h"
#include "file.h"
#include "ins.h"
#include "ledit.h"
#include "screen.h"
#include "window.h"
#include "display.h"
#include "search.h"
#include "misc.h"
#include "replace.h"
#include "paste.h"
#include "sysdep.h"
#include "cmds.h"
#include "text.h"
#include "slang.h"



#define JED_PROMPT "JED>"
Buffer *MiniBuffer;
Buffer *Last_Buffer;
char *Token_Pointer;
Buffer *The_MiniBuffer;     /* this never gets deleted since we may need it */
Window_Type *The_MiniWindow;

extern void get_cwd(char *);

User_Variable_Type User_Vars =
  {
      78,				/* wrap-column */
      8,				/* tabs */
      3,				/* c_indent */
      2,				/* c_brace */
     300,				/* max_hits */
  };

char Last_Search_Str[132];

int (*complete_open)(char *);
int (*complete_next)(char *);

static int Buf_List_Len;
Buffer *Buf_Context;

int next_bufflist(char *buf)
{
   Buffer *this;
   while (1)
     {
	this = Buf_Context;
	if (this == MiniBuffer) return(0);
	Buf_Context = Buf_Context->next;
	if (!Buf_List_Len || !strncmp(buf, this->name, Buf_List_Len)) 
	  {
	     strcpy(buf, this->name);
	     return(1);
	  }
     }
}


int open_bufflist(char *buf)
{
   if ((Buf_Context = MiniBuffer) == NULL) return(0);
   Buf_Context = Buf_Context->next;
   Buf_List_Len = strlen(buf);
   return next_bufflist(buf);
}



char *what_buffer()
{
   return(CBuf->name);
}

int bufferp(char *name)
{
   if (NULL == find_buffer(name)) return(0); else return(1);
}



int insert_buffer_name(char *name)
{
   Buffer *buf;

   if (NULL != (buf = find_buffer(name)))
     {
	insert_buffer(buf);
	return(1);
     }
   else msg_error("Unable to find buffer.");
   return(0);
}


int replace_cmd(char *old, char *new)
{
   char msgstr[132];
   
   if (*old == 0)
     {
	msg_error("replace what?");
	return(0);
     }
    
   sprintf(msgstr,"Replace \"%s\" with \"%s\"? (y/n/!/q)", old, new);
   push_spot();
   backwchars(1);
   if (search(old, 1, 0))
     {
	while(1)
	  {
	     message(msgstr);
	     Window->trashed = 1;
	     update((Line *) NULL);
	     fflush(stdout);
	     switch(getkey())
	       {
		  case 'Y': case 'y': replace_next(old, new); break;
		  case '!': while(replace_next(old, new)); break;
		  case 'q': case 'Q': return(1);
		  case 'N': case 'n': forwchars(1); break;
		  default: beep(); return(1);
	       }	     
	     if (!search(old, 1, 0)) break;
	  }
     }
   else pop_spot();
   
   message("Done.");
   return(1);
}

int replace()
{
   static char old[80], new[80];
   
   complete_open = NULL;
   
   if (!read_from_minibuffer("Replace:", 0, old)) return(0);
   if (!read_from_minibuffer("With:", 0, new)) return(0);
   
   replace_cmd(old, new);
   return(1);
}


int kill_buffer_cmd(char *name)
{
   Buffer *buf, *kill_buf, *scratch;
   Window_Type *w;
   int kill;

   if (!strcmp(name, "*scratch*"))
     {
	erase_buffer();
	return(1);
     }
   scratch = find_buffer("*scratch*");
   if (NULL == scratch)
     {
	scratch = make_buffer();
	strcpy(scratch->name, "*scratch*");
	strcpy(scratch->dir, CBuf->dir);
	*scratch->file = 0;
     }

   if (NULL == (kill_buf = find_buffer(name)))
     {
	msg_error("Buffer does not exist.");
	return(0);
     }
   
   kill = 1;
   if (kill_buf->flags & BUFFER_TRASHED)
     {
	message("Buffer modified. Kill, Save First, Abort: [KSA]");
	flush_message();
	switch (getkey())
	  {
	     case 'k': case 'K': kill = 1; break;
	     case 's': case 'S': kill = 2; break;
	     default: msg_error("Aborted."); return(0);
	  }
     }
   
   if (kill == 2)
     {
	buf = CBuf;
	switch_to_buffer(kill_buf);
	write_buffer();
	switch_to_buffer(buf);
	if (*Error_Buffer) return(0);
     }
   
   
   buf = kill_buf->prev;
   while(buf != kill_buf)
     {
	if ((buf != scratch) && (*buf->name != ' ')) break;
	buf = buf->prev;
     }
   
   if (buf == kill_buf) buf = scratch;
   
   if (kill_buf == CBuf)
     {
	switch_to_buffer(buf);
	window_buffer(CBuf);
     }
   
   /* search through windows looking for the buffer and replace it */
   w = Window;
   Window = Window->next;
   while(w != Window)
     {
	if (kill_buf == Window->buffer)
	  {
	     window_buffer(buf);  /* I should choose a different one */
	  }
	Window = Window->next;
     }
   if (kill_buf == Last_Buffer) Last_Buffer = NULL;
   delete_buffer(kill_buf);
   return(0);
}


int write_buffer_cmd(char *file)
{
#ifdef VMS
   char *ff;
#endif   
   char dir[255], *f;
   char msgstr[132];
   int n;
   
   if ((*file == 0) 
       || (f = expand_filename(file), (*f == 0) && (*CBuf->file == 0)))
     {
	msg_error("Filename Required.");
	return(0);
     }
   strncpy(dir, file, (int) (f - file));
   dir[(int) (f - file)] = 0;
   
   if (*f == 0) f = CBuf->file;
   
   n = write_file_with_backup(dir, f);
   if (n >= 0)
     {
#ifdef VMS
	ff = f; while (*ff) if (*ff == ';') *ff = 0; else ff++;
#endif	
	sprintf(msgstr,"Wrote %d lines to %s%s", n, dir, f);
	message(msgstr);
	CBuf->flags &= ~BUFFER_TRASHED;
	CBuf->flags |= AUTO_SAVE_BUFFER;
	CBuf->hits = 0;
	visit_file(dir, f);
	return(1);
     }
   else
     {
	sprintf(msgstr,"Error writing file %s%s", dir, f);
	msg_error(msgstr);
	return(0);
     }
}

#ifdef msdos
int show_memory()
{
   char buf[40];
   sprintf(buf, "%lu bytes free.", farcoreleft());
   message(buf);
   return(0);
}
#endif


int set_buffer(char *name)
{
    Buffer *buf;
    
    if ((name == NULL) || (*name == 0)) return(0);
    /* Last_Buffer = CBuf; */
    
    if (NULL == (buf = find_buffer(name)))
      {
	 buf = make_buffer();
	 strcpy(buf->dir, CBuf->dir);
	 *buf->file = 0;	  
	 switch_to_buffer(buf);
	 uniquely_name_buffer(name);
      }
    else switch_to_buffer(buf);
    return(1);
}


int insert_file_cmd()
{
    char file[255], *f;
    
   complete_open = sys_findfirst;
   complete_next = sys_findnext;

   if (*CBuf->dir == 0) get_cwd(CBuf->dir);
   strcpy(file, CBuf->dir);
   if (!read_from_minibuffer("Insert file:", 0, file)) return(0);
   
   f = expand_filename(file);
   if ((*file == 0) || (*f == 0))
     {
	msg_error("Expecting filename.");
	return(1);
     }
   
   if (insert_file(file) < 0) msg_error("Error inserting file.");
   return(1);
}


int find_file_cmd(char *file)
{
   char *f;
   char msgstr[132];
   Buffer *buf;
   int n;

   f = expand_filename(file);
   if ((*file == 0) || (*f == 0))
     {
	msg_error("Expecting filename.");
	return(0);
     }
   
   /* search for the file in current buffer list */

   if (NULL != (buf = find_file_buffer(file)))
     {
	if (buf == CBuf) return(0);
	Last_Buffer = CBuf;
	switch_to_buffer(buf);
	window_buffer(buf);
	return(1);
     }
   
   Last_Buffer = CBuf;
   buf = make_buffer();
   switch_to_buffer(buf);
   
   strcpy(CBuf->file, f);
   strncpy(CBuf->dir, file, (int)(f - file));
   (CBuf->dir)[(int)(f - file)] = 0;
   
   uniquely_name_buffer(f);
   n = read_file(file);
   if (n != -1)
     {
	sprintf(msgstr,"%d lines read", n);
	message(msgstr);
     }
   else message("New File.");
   CLine = CBuf->beg;
   if (CLine == NULL) make_line(25);
   set_file_modes();
   window_buffer(CBuf);
   
   return(1);
}


void set_mode_cmd(char *mode)
{
   *CBuf->mode_str = 0;
   if (!strcmp("wrap", mode)) CBuf->modes = WRAP_MODE;
   else if (!strcmp("cmode", mode)) CBuf->modes = C_MODE;
   else if (!strcmp("nomode", mode)) CBuf->modes = NO_MODE;
   else
     {
	CBuf->modes = LANG_MODE;
	strncpy(CBuf->mode_str, mode, 12);
	CBuf->mode_str[12] = 0;
     }
   /*  else msg_error("Undefined Mode"); */
}


int c_mode_cmd()
{
    CBuf->modes = C_MODE;
    return(0);
}

int do_search_cmd(int direct)
{
   
   char *s, *p, tmp[80], *str = Last_Search_Str;
    
   complete_open = NULL;

    if (direct == 1) p = "Search:"; else p = "Search backward:";
    if (*str == 0) s = NULL; else s = str;
    *tmp = 0;
    if (!read_from_minibuffer(p, s, tmp)) return(0);
    
    if (! *tmp) return(0);
    strcpy(str, tmp);
    if (direct == 1) forwchars(1);/*  else backwchars(1); */
    if (search(str, direct, 0)) return(1);
    msg_error("Not Found!");
    return(0);
}


int search_forward_cmd()
{
   return do_search_cmd(1);
}

int search_backward_cmd()
{
   return do_search_cmd(-1);
}


/* create a minibuffer with window and switch to it. */
void create_minibuffer(void)
{
   Window_Type *w;
   w = The_MiniWindow;
   MiniBuffer = The_MiniBuffer;
   w->next = Window->next;
   Window->next = w;
   Mini_Info.action_window = Window;
   other_window();    /* now in minibuffer window */
   window_buffer(MiniBuffer);	
   MiniBuffer->keymap = Mini_Map;
   switch_to_buffer(MiniBuffer);
   eob();
   bol();
   CLine->len = 0;
}

/* evaluate command in minibuffer and leave */
int exit_minibuffer()
{
   if (IS_MINIBUFFER) Exit_From_MiniBuffer = 1;
   return(1);
}

   
   
/* return 1 if minibuffer already exists otherwise returns 0 */
int select_minibuffer()
{
    Window_Type *w;
    
    /* Try to find active minibuffer and go there */    
    w = Window;
    while (MiniBuffer != NULL)
      {
	 if (Window->top == Screen_Height) return(1);
	 other_window();
	 if (w == Window) exit_error("Internal Error:  no window!");
      }
    
    /* switchs to minibuffer too */
    create_minibuffer();
    return(0);
}


/* if cmd != NULL, insert it into minibuffer and then send the result to
   the appropriate routine. */
   
int ledit(void)
{
   char tmp[132];
   *tmp = 0;   
   complete_open = NULL;
   if (!read_from_minibuffer(JED_PROMPT, 0, tmp)) return(0);
   Lang_Error = 0;
   interpret(tmp);
   if ((Lang_Error == -1) && KeyBoard_Quit)
     {
	msg_error("Quit!");
     }
   Lang_Error = 0;
   
   return(1);
}


int find_file()
{
   char tmp[255];
   if (*CBuf->dir == 0) get_cwd(CBuf->dir);
   strcpy(tmp, CBuf->dir);
   
   complete_open = sys_findfirst;
   complete_next = sys_findnext;

   if (!read_from_minibuffer("Find file:", 0, tmp)) return(0);
   find_file_cmd(tmp);
   return(0);
}

int write_buffer()
{
    char tmp[255];
    
    if (*CBuf->dir == 0) get_cwd(CBuf->dir);
    strcpy(tmp, CBuf->dir);
    if (!read_from_minibuffer("write file:", 0, tmp)) return(0);
    write_buffer_cmd(tmp);
    return(0);
}

void switch_to_buffer_cmd(char *name)
{
   set_buffer(name);
   window_buffer(CBuf);
}


int get_buffer()
{
    char tmp[255];
   *tmp = 0;
   if (Last_Buffer == NULL) Last_Buffer = CBuf;
   
   complete_open = open_bufflist;
   complete_next = next_bufflist;
   
   if (!read_from_minibuffer("Switch to buffer:", Last_Buffer->name, tmp)) return(0);
   switch_to_buffer_cmd(tmp);
   return(1);
}

    


int read_integer_from_minibuffer(char *prompt, int *x)
{
   char tmp[132];
   *tmp = 0;
   complete_open = NULL;
   if (!read_from_minibuffer(prompt, 0, tmp)) return(0);
   if (1 != sscanf(tmp, "%d", x)) 
     {
	msg_error("Expecting Integer.");
	return(0);
     }
   
   return(1);
}


int goto_line_cmd()
{
   int x;
   
   if (!read_integer_from_minibuffer("Goto line:", &x)) return(0);
   goto_line(x);
   return(1);
}



int goto_column_cmd()
{
   int x;   
   if (!read_integer_from_minibuffer("Goto column:", &x)) return(0);
   goto_column(x);
   return(1);
}


int kill_buffer()
{
   char tmp[255];
   *tmp = 0;
   
   complete_open = open_bufflist;
   complete_next = next_bufflist;
   if (!read_from_minibuffer("Kill buffer:", (char *) CBuf->name, tmp)) return(0);
   kill_buffer_cmd(tmp);
   return(1);
}

int evaluate_cmd()
{
    return(!ledit());
}

void insert_string(char *s)
{
   ins_chars((unsigned char *) s, strlen(s));
}

void set_key(char *function, char *key)
{   
   define_key(key, function, Global_Map);
}

void unset_key(char *key)
{
   undefine_key(key, Global_Map);
}

void unset_key_in_keymap(char *key, char *map)
{
   Key_Type *kmap;
   
   if (NULL == (kmap = find_keymap(map)))
     {
	msg_error("Unknown keymap.");
	return;
     }
   
   undefine_key(key, kmap);
}

int keymap_p(char *name)
{
   return ! (NULL == find_keymap(name));
}


void set_key_in_keymap(char *f, char *key, char *map)
{
   Key_Type *kmap;
   
   if (NULL == (kmap = find_keymap(map)))
     {
	msg_error("Unknown keymap.");
	return;
     }
   
   define_key(key, f, kmap);
}


void pop_to_buffer(char *name)
{
   Window_Type *w, *action, *use_this;

   
   use_this = NULL;
   if (MiniBuffer != NULL)
     {
	action = Mini_Info.action_window;
     }
   else action = NULL;
   
   w = Window;
   /* find a window to use */
   do 
     {
	if (w->top != Screen_Height)
	  {
	     if (action != NULL) 
	       {
		  if (w != action) use_this = w;
	       }
	     else if (w != Window) use_this = w;
	     
	     if (!strcmp(w->buffer->name, name)) 
	       {
		  use_this = w;
		  break;
	       }	     
	  }
	w = w->next;
     }
   while (w != Window);
   
   if (use_this != NULL) 
     {
	while(Window != use_this) other_window();
     }
   else
     {
	if (action != NULL) while(Window != action) other_window();
	split_window();
     }
   
   switch_to_buffer_cmd(name);
}

/* wrapper functions for the editor/ language interface */

void goto_column_wrapper(int *n)
{
   goto_column(*n);
}

int prevline_wrapper(int *n)
{
   return(prevline(*n));
}

int nextline_wrapper(int *n)
{
   return(nextline(*n));
}

int backwchars_wrapper(int *n)
{
   return(backwchars(*n));
}

int forwchars_wrapper(int *n)
{
   return(forwchars(*n));
}

void goto_line_wrapper(int *n)
{
   goto_line(*n);
}


/* return number of windows */
int num_windows()
{
   Window_Type *w;
   int i = 0;
   
   w = Window->next;
   while (i++, w != Window) w = w->next;
   return(i);
}



void flush_message(void)
{
   do_dialog(Message_Buffer);
   *Message_Buffer = 0;
   Window->trashed = 1;
}



#ifndef msdos
#ifndef VMS

int jed_mail_send()
{
   FILE *pp;
   int n;
   char subject[80], to[80], cmd[160];
     
   push_spot();
   bob();
   push_mark();
   if (!search_forward("---text follows this line---"))
     {
	msg_error("Failed to find text divider.");
	return(0);
     }
   eol();
   narrow_to_region();
   if (!search("To:", -1, 0))
     {
	msg_error("No To: field.");
	widen();
	pop_spot();
	return(0);
     }
   forwchars(3);
   n = CLine->len - Point;
   strncpy(to, (char *) (CLine->data + Point), n);
   to[n - 1] = 0;
   
   bob();
   if (search("Subject:", 1, 0))
     {
	forwchars(8);
	n = CLine->len - Point;
	strncpy(subject, (char *) (CLine->data + Point), n);
	subject[n - 1] = 0;
     }
   else *subject = 0;
   
   sprintf(cmd, "mail -s \"%s\" %s", subject, to);
   
   if (NULL == (pp = popen(cmd, "w")))
     {
	msg_error("Unable to open mail process.");
	widen();
	pop_spot();
	return(0);
     }

   eob();
   widen();
   message("mailing...");
   flush_message();
   forwchars(1);
   push_mark();
   eob();
   narrow_to_region();
   
   write_filepointer(pp);
   pclose(pp);
   message("mailing...done");
   widen();
   pop_spot();
   CBuf->flags = CBuf->flags & ~BUFFER_TRASHED;
   return(1);
}

int jed_mail()
{
   int exists;
   
   exists = bufferp("*mail*");
   pop_to_buffer("*mail*");
   if (!exists)
     {
	CBuf->modes = TEXT_MODE;
	insert_string("To: "); newline();
	insert_string("Subject: "); newline();
	insert_string("---text follows this line---"); newline();
	bob();
	eol();
     }
   
   return(1);
}
#endif
#endif


int load_file(char *file)
{
   FILE *fp;
   char buf[132], count, *lib, libfile[255];
   
   if (NULL == (fp = fopen(file, "r"))) 
     {
	if (NULL != (lib = getenv("JED_LIBRARY")))
	  {
	     strcpy(libfile, lib);
	     fixup_dir(libfile);
	     strcat(libfile, file);
	     file = libfile;
	  }
	
	if (NULL == (fp = fopen(file, "r"))) 
	  {
	     msg_error("Unable to load file.");
	     return(0);
	  }
     }
   count = 0;
   sprintf(buf, "loading %s...", file);
   message(buf);
   flush_message();
   while(NULL != fgets(buf, 131, fp))
     {
	count++;
	interpret(buf);
	if (Lang_Error || *Error_Buffer)
	  {
	     Lang_Error = 0;
	     break;	     
	  }	
     }
   if (! *Error_Buffer)
     {
	sprintf(buf, "loading %s...done", file);
	message(buf);
	do_dialog(Message_Buffer);
     }   

   fclose(fp);
   return(1);
}

void load_buffer()
{
   char buf[132];
   Line *l;
   unsigned char *offset;

   l = CBuf->beg;
   do 
     {
	/* warning--- if the line is bigger than 132 then we are dead.
	   Also I should be able to send only the data without the strncpy
	   but the last line may not be null terminated. */
	   
	strncpy(buf, (char *) l->data, l->len);  buf[l->len] = 0;
	offset = (unsigned char *) interpret(buf);
	if (Lang_Error || *Error_Buffer)
	  {
	     CLine = l;
	     Point = (int) (offset - (unsigned char *)buf);
	     (void) skip_whitespace();
	     break;	     
	  }
	l = l->next;
     }
   while(l != NULL);
   return;
}

/* pop up a list of buffers */
int list_buffers()
{
   Buffer *b;
   char dsc[132];
   int mini;
   
   mini = IS_MINIBUFFER;
   
   pop_to_buffer("*buffers*");
   b = CBuf;
   do
     {
	sprintf(dsc, "%s\t\t\t%s%s\n",b->name,b->dir,b->file);
	insert_string(dsc);
	b = b->next;
     }
   while(b != CBuf);
   bob();
   CBuf->flags = 0;
   
   if (mini) while(!IS_MINIBUFFER) other_window();
   return(1);
}


int show_key()
{
   char msg[80], *str;
   
   message("Show key:");
   update((Line *) NULL);
   fflush(stdout);
   str = find_key();
   if (str != NULL)
     {
	sprintf(msg, "Key runs cmd \"%s\".", str);
     }
   else strcpy(msg, "Key is undefined.");
   message(msg);
   return(1);
}

int what_char()
{
   if (eobp()) return(0);
   return( (int) *(CLine->data + Point) );
}

void update_cmd()
{
   Window->trashed = 1;
   touch_window();
   update((Line *) NULL);
}

void call_cmd(char *str)
{
   int (*fp)();
   
   if (NULL == (fp = (int (*)()) find_function(str)))
     {
	msg_error("Function does not exist!");
     }
   else (void) (*fp)();
}

void copy_region_cmd(char *name)
{
   Buffer *buf;
   
   if (NULL != (buf = find_buffer(name)))
     {
	copy_region_to_buffer(buf);
     }
   else msg_error("Unable to find buffer.");
}

void screen_w80()
{
   narrow_width();
   change_screen_size(80, Screen_Height);
}

void screen_w132()
{
   wide_width();
   change_screen_size(132, Screen_Height);
}

/* only works on the current line */
char *buffer_substring(void)
{
   static char string[255];
   unsigned char *tmp, *p1, *p2;
   
   if (CBuf->marks == NULL) 
     {
	p1 = CLine->data + Point;
	p2 = CLine->data + CLine->len;
     }
   else
     {
	p1 = CLine->data + CBuf->marks->point;
	p2 = CLine->data + Point;
	if (p2 < p1) 
	  {
	     tmp = p1; p1 = p2; p2 = tmp;
	  }
	push_spot(); pop_mark(); pop_spot();
     }
   if (p2 - p1 > 254) p2 = p1 + 254;
   tmp = (unsigned char *) string;
   while (p1 < p2) *tmp++ = *p1++;
   *tmp = 0;
   return(string);
}



int mini_complete()
{
   char *buf, *pl, *pb;
   char last[255];
   static char prev[255];
   int n;
   static int flag = 0;  /* when flag goes 0, we call open */
   
   if (complete_open == NULL) return ins_char_cmd();
   
   Point = 0;
   push_mark();
   eol();
   buf = buffer_substring();
   
   
   if ((Last_Command_Char == ' ') && (Last_Key_Function == (VOID *) mini_complete))
     {
	if (!flag || !(flag = (*complete_next)(buf)))
	  {
	     strcpy(buf, prev);
	     flag = (*complete_open)(buf);
	  }
	
	strcpy(last, buf);
	n = -1;
     }
   else 
     {
	n = 0;
	strcpy(prev, buf);  /* save this search context */
     }
   
	
   if (!n && (flag = (*complete_open)(buf), flag))
     {
	strcpy(last, buf);
	do
	  {
	     n++;
	     pl = last;
	     pb = buf;
	     while (*pl && (*pl == *pb)) pl++, pb++;
	     if (*pl) *pl = 0;
	  }
	while(flag = (*complete_next)(buf), flag);
     }
   
   if (n)
     {
	Point = 0;
	push_mark();
	eol();
	delete_region();
	insert_string(last);
	if ((n == 1) && (Last_Key_Function == (VOID *) mini_complete))
	  message("[Sole Completion.]");
     }
   else msg_error("No Match!");
   
   return(1);
}

int markp(void)
{
   return (CBuf->marks != NULL);
}

int dup_mark(void)
{
   if (CBuf->marks == NULL) return(0);
   
   push_spot(); 
   CLine = CBuf->marks->line; Point = CBuf->marks->point;
   push_mark();
   pop_spot();
   return(1);
}


char *mini_read(char *prompt, char *def)
{
   static char buf[255];

   *buf = 0;
   complete_open = NULL;
   if (!read_from_minibuffer(prompt, def, buf)) return "";
   return(buf);
}

   

void init_minibuffer()
{  
   Buffer *tmp;
   
   tmp = CBuf;
   The_MiniBuffer = make_buffer();
   The_MiniBuffer->modes = 0;
   strcpy(The_MiniBuffer->name, " <mini>");
   switch_to_buffer(The_MiniBuffer);  /* do some initializing */
   remake_line(128);
   switch_to_buffer(tmp);
   The_MiniWindow = create_window(Screen_Height, 1, 1, 80);

   init_lang((int (*)()) msg_error);
#ifdef msdos   
   add_intrinsic("coreleft", (long) show_memory,VOID_TYPE, 0);
#else
#ifndef VMS   
   add_intrinsic("mail", (long) jed_mail, VOID_TYPE, 0);
   add_intrinsic("send", (long) jed_mail_send, VOID_TYPE, 0);
#endif
#endif
   add_intrinsic("error", (long) msg_error, VOID_TYPE, 1);
   add_intrinsic("read_mini", (long) mini_read, STRING_TYPE, 2);
   add_intrinsic("bufsubstr", (long) buffer_substring, STRING_TYPE, 0);
   add_intrinsic("call", (long) call_cmd, VOID_TYPE, 1);
   add_intrinsic("self_ins", (long) ins_char_cmd, VOID_TYPE, 0);
   add_intrinsic("flush", (long) flush_message, VOID_TYPE, 0);
   add_intrinsic("setmode", (long) set_mode_cmd, VOID_TYPE, 1);
   add_intrinsic("skip_chars", (long) skip_chars, VOID_TYPE, 1);
   add_intrinsic("skip_white", (long) skip_whitespace, VOID_TYPE, 0);
   add_intrinsic("looking_at", (long) looking_at, INT_TYPE, 1);
   add_intrinsic("what_char", (long) what_char, INT_TYPE, 0);
   
   add_intrinsic("showkey", (long) show_key, VOID_TYPE, 0);
   add_intrinsic("erase_buffer", (long) erase_buffer, VOID_TYPE, 0);
   add_intrinsic("evalfile", (long) load_file, INT_TYPE, 1);
   add_intrinsic("evalbuffer", (long) load_buffer, VOID_TYPE ,0);
   add_intrinsic("whitespace", (long) insert_whitespace,VOID_TYPE, 1);
   add_intrinsic("onewindow", (long) one_window,VOID_TYPE, 0);
   add_intrinsic("otherwindow", (long) other_window,VOID_TYPE, 0);
   add_intrinsic("splitwindow", (long) split_window,VOID_TYPE, 0);
   add_intrinsic("enlargewin", (long) enlarge_window,VOID_TYPE, 0);
   add_intrinsic("pageup", (long) pageup_cmd,VOID_TYPE, 0);
   add_intrinsic("pagedown", (long) pagedown_cmd,VOID_TYPE, 0);
   add_intrinsic("nwindows", (long) num_windows,INT_TYPE, 0);
   
   add_intrinsic("bol", (long) bol,VOID_TYPE, 0);
   add_intrinsic("bob", (long) bob,VOID_TYPE, 0);
   add_intrinsic("bobp", (long) bobp,INT_TYPE, 0);
   add_intrinsic("bolp", (long) bolp,INT_TYPE, 0);
   add_intrinsic("eol", (long) eol,VOID_TYPE, 0);
   add_intrinsic("eob", (long) eob,VOID_TYPE, 0);
   add_intrinsic("eolp", (long) eolp,INT_TYPE, 0);
   add_intrinsic("eobp", (long) eobp,INT_TYPE, 0);
   add_intrinsic("insert", (long) insert_string,VOID_TYPE, 1);
   add_intrinsic("del", (long) del,VOID_TYPE, 0);
   add_intrinsic("trim", (long) trim_whitespace,VOID_TYPE, 0);
   add_intrinsic("up", (long) prevline_wrapper,INT_TYPE, 1);
   add_intrinsic("down", (long) nextline_wrapper,INT_TYPE, 1);
   add_intrinsic("left", (long) backwchars_wrapper,INT_TYPE, 1);
   add_intrinsic("right", (long) forwchars_wrapper,INT_TYPE, 1);
   add_intrinsic("what_column", (long) calculate_column,INT_TYPE, 0);
   add_intrinsic("whatline", (long) what_line,INT_TYPE, 0);
   add_intrinsic("push_spot", (long) push_spot,VOID_TYPE, 0);
   add_intrinsic("pop_spot", (long) pop_spot,VOID_TYPE, 0);
   add_intrinsic("push_mark", (long) push_mark,VOID_TYPE, 0);
   add_intrinsic("goto_column", (long) goto_column_wrapper,VOID_TYPE, 1);
   add_intrinsic("goto_line", (long) goto_line_wrapper,VOID_TYPE, 1);
   
   
   add_intrinsic("keymap_p", (long)  keymap_p, INT_TYPE, 1);
   add_intrinsic("use_keymap", (long)  use_keymap, VOID_TYPE, 1);
   add_intrinsic("make_keymap", (long) create_keymap, VOID_TYPE, 1);
   add_intrinsic("definekey", (long) set_key_in_keymap, VOID_TYPE, 3);
   add_intrinsic("setkey", (long) set_key, VOID_TYPE, 2);
   add_intrinsic("undefinekey", (long) unset_key_in_keymap, VOID_TYPE, 2);
   add_intrinsic("unsetkey", (long) unset_key, VOID_TYPE, 1);
   add_intrinsic("ffind", (long) forward_search_line, INT_TYPE, 1);
   add_intrinsic("fsearch", (long) search_forward, INT_TYPE, 1);
   add_intrinsic("bfind", (long) backward_search_line, INT_TYPE, 1);
   add_intrinsic("bsearch", (long) search_backward, INT_TYPE, 1);
   add_intrinsic("narrow", (long) narrow_to_region, VOID_TYPE, 0);
   add_intrinsic("widen", (long) widen, VOID_TYPE, 0);
   add_intrinsic("setbuf", (long) set_buffer, VOID_TYPE, 1);
   add_intrinsic("sw2buf", (long) switch_to_buffer_cmd, VOID_TYPE, 1);
   add_intrinsic("pop2buf", (long) pop_to_buffer, VOID_TYPE, 1);
   add_intrinsic("delbuf", (long) kill_buffer_cmd, VOID_TYPE, 1);
   add_intrinsic("insbuf", (long) insert_buffer_name, VOID_TYPE, 1);
   add_intrinsic("whatbuf", (long) what_buffer, STRING_TYPE, 0);
   add_intrinsic("bufferp", (long) bufferp, INT_TYPE, 1);
   add_intrinsic("message", (long) message, VOID_TYPE, 1);
   add_intrinsic("time", (long) get_time, STRING_TYPE, 0);
   add_intrinsic("exit_jed", (long) exit_jed, VOID_TYPE, 0);
   add_intrinsic("find_file", (long) find_file_cmd, INT_TYPE, 1);
   add_intrinsic("write_buffer", (long) write_buffer_cmd, INT_TYPE, 1);
   add_intrinsic("indent_line", (long) indent_line, VOID_TYPE, 0);
   add_intrinsic("copy_rect", (long) copy_rectangle, VOID_TYPE, 0);
   add_intrinsic("open_rect", (long) open_rectangle, VOID_TYPE, 0);
   add_intrinsic("blank_rect", (long) blank_rectangle, VOID_TYPE, 0);
   add_intrinsic("kill_rect", (long) kill_rectangle, VOID_TYPE, 0);
   add_intrinsic("insert_rect", (long) insert_rectangle, VOID_TYPE, 0);
   add_intrinsic("update", (long) update_cmd, VOID_TYPE, 0);
   
   
   add_intrinsic("del_region", (long) delete_region, VOID_TYPE, 0);
   add_intrinsic("copy_region", (long) copy_region_cmd, VOID_TYPE, 1);
   
   add_intrinsic("w132", (long) screen_w132, VOID_TYPE, 0);
   add_intrinsic("w80", (long) screen_w80, VOID_TYPE, 0);
   add_intrinsic("markp", (long) markp, INT_TYPE, 0);
   add_intrinsic("dupmark", (long) dup_mark, INT_TYPE, 0);
   
   add_intrinsic("skip_word", (long) forward_word, VOID_TYPE, 0);
   add_intrinsic("bskip_word", (long) backward_word, VOID_TYPE, 0);
   add_intrinsic("xform_reg", (long) transform_region, VOID_TYPE, 1);
   
   
   add_variable("LAST_SEARCH", (long) Last_Search_Str, STRING_TYPE);
   add_variable("TAB", (long) &User_Vars.tab, INT_TYPE);
   add_variable("WRAP", (long) &User_Vars.wrap_column, INT_TYPE);
   add_variable("C_INDENT", (long) &User_Vars.c_indent, INT_TYPE);
   add_variable("C_BRACE", (long) &User_Vars.c_brace, INT_TYPE);
   add_variable("LAST_CHAR", (long) &Last_Command_Char, INT_TYPE);
   add_variable("MAX_HITS", (long) &User_Vars.max_hits, INT_TYPE);
   add_variable("CASE_SEARCH", (long) &Case_Sensitive, INT_TYPE);
   /* add_variable("SCRN_HEIGHT", (long) &Screen_Height, INT_TYPE);
      add_variable("SCRN_WIDTH", (long) &Screen_Width, INT_TYPE); */
}


