/*
 *  Copyright (c) 1992 John E. Davis  (davis@amy.tch.harvard.edu)
 *  All Rights Reserved.
 */
    
/*  This file is the interface to system specific files */

#include <stdio.h>
#include "buffer.h"
#include "sysdep.h"
#include "misc.h"

int Screen_Height = 0;
int Screen_Width = 0;

#ifdef msdos
#include "ibmpc.c"
#else
#ifdef VMS
#include "vms.c"
#else
#ifdef BSD
#include "bsd.c"
#else
#include "sys5.c"
#endif /* bsd */
#endif /* vms */
#endif /* ibm */

#ifdef unix
#include <sys/stat.h>
#include <errno.h>
#endif

int INPUT_BUFFER_LEN = 0;
char INPUT_BUFFER[255];

char getkey()
{
   int i, imax;
   char ch;
   if (!INPUT_BUFFER_LEN) return(sys_getkey());
   ch = INPUT_BUFFER[0];
   INPUT_BUFFER_LEN--;
   imax = INPUT_BUFFER_LEN;
   for (i = 0; i < imax; i++)
     {
	INPUT_BUFFER[i] = INPUT_BUFFER[i + 1];
	imax = INPUT_BUFFER_LEN;
     }
   return(ch);
}

void ungetkey(char ch)
{
   int i;
   for (i = 0; i < INPUT_BUFFER_LEN; i++)
     {
	INPUT_BUFFER[i+1] = INPUT_BUFFER[i];
     }
   INPUT_BUFFER[0] = ch;
   INPUT_BUFFER_LEN++;
}

void flush_input()
{
   int quit = KeyBoard_Quit;
   INPUT_BUFFER_LEN = 0;
   KeyBoard_Quit = 0;
#ifdef msdos
   while (input_pending(2,0)) if (!getkey()) getkey();
#else
   while (input_pending(2,0)) getkey();
#endif   
   KeyBoard_Quit = quit;
}


#include <time.h>

char *get_time()
{ 
    time_t clock;
    char *the_time;

    clock = time((time_t *) 0);
    the_time = ctime(&clock); 
    /* returns the form Sun Sep 16 01:03:52 1985\n\0 */
    the_time[24] = '\0';
    return(the_time);
}

#ifndef msdos
int sys_chmod(char *file, int what, int *mode)
{
#ifdef VMS
   return(0);
#else
   
   struct stat buf;
   char *ms, *es;
   
   if (what == 1) 
     {
	chmod(file, *mode);
	return(0);
     }
   
   if (stat(file, &buf) < 0)
     {
	ms = es = NULL;
	switch (errno)
	  {
	     case EACCES: es = "Access denied."; break;
	     case ENOENT: ms = "File does not exist."; break;
	     default: es = "stat: unknown error."; break;
	  }
	if (ms == NULL) 
	  {
	     msg_error(es); return(-1);
	  }	
	else message(ms); 
	return(0);
     }
   
   *mode = buf.st_mode;
   if (*mode & S_IFDIR)
     {
	msg_error("File is a directory.");
	return (-1);
     }
   *mode &= 0777;
   return(1);
#endif /* vms */
}
#endif /* msdos */

#ifdef unix
#include <sys/dir.h>

static char Unix_Dir[255];
static char Unix_File[255];
static int File_Len;
static DIR *Dirp;

int sys_findnext(char *file)
{
   struct direct *dp;
   
   while (1)
     {
	if (NULL == (dp = readdir(Dirp))) return(0);
	dp->d_name[dp->d_namlen] = 0;
	if (!strncmp(Unix_File, dp->d_name, File_Len)) break;
     }
   strcpy(file, Unix_Dir); strcat(file, dp->d_name);
   return(1);
}


int sys_findfirst(char *file)
{
   char *f;
   
   f = expand_filename(file);
   strcpy (Unix_Dir, file);
   strcpy (Unix_File, f);
   File_Len = strlen(f);
   
   /* if (f != file) f--; */
   Unix_Dir[(int) (f - file)] = 0;
   
   if (Dirp != NULL) closedir(Dirp);
   
   if (NULL == (Dirp = (DIR *) opendir(Unix_Dir))) return(0);
   return sys_findnext(file);
}
#endif
